
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdint.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <fcntl.h>
#include <string.h>
#include <sys/ioctl.h>
#include <linux/types.h>
#include <linux/spi/spidev.h>
#include <linux/i2c-dev.h>

#include "lcd.h"
#include "bwi2c.h"
#include "report.h"

#define DEFAULT_WIDTH   LCD_DEFAULT_WIDTH
#define DEFAULT_HEIGHT    LCD_DEFAULT_HEIGHT
#define DEFAULT_CELLWIDTH LCD_DEFAULT_CELLWIDTH
#define DEFAULT_CELLHEIGHT  LCD_DEFAULT_CELLHEIGHT
#define DEFAULT_CONTRAST  500
#define DEFAULT_BRIGHTNESS  750
#define DEFAULT_DEVICE    "/dev/i2c-0"

/** private data for the \c debug driver */
typedef struct bwi2c_private_data {
  unsigned char *framebuf;  /** frame buffer */
  unsigned char *lcdcontents;  /** buffer for old lcd contents */
  int fd;  /**< device file handle */
  int port;     
  int width;    /**< display width in characters */
  int height;   /**< display height in characters */
  int cellwidth;    /**< character cell width */
  int cellheight;   /**< character cell height */
  int contrast;   /**< current contrast */
  int brightness;   /**< current brightness (for backlight on) */
} PrivateData;

#define ARRAY_SIZE(a) (sizeof(a) / sizeof((a)[0]))

/* Vars for the server core */
MODULE_EXPORT char *api_version = API_VERSION;
MODULE_EXPORT int stay_in_foreground = 1;
MODULE_EXPORT int supports_multiple = 0;
MODULE_EXPORT char *symbol_prefix = "bwi2c_";


/**
 * Initialize the driver. (Required)
 * \param drvthis  Pointer to driver structure.
 * \retval 0       Success.
 * \retval <0      Error.
 */
MODULE_EXPORT int
bwi2c_init(Driver *drvthis)
{
  PrivateData *p;
  char device[256] = DEFAULT_DEVICE;
  const char *s;
  //int addr;

  /* Allocate and store private data */
  p = (PrivateData *) calloc(1, sizeof(PrivateData));
  if (p == NULL) {
    return -1;
  }

  if (drvthis->store_private_ptr(drvthis, p)) {
    return -1;
  }

  /* Get and parse size */
  s = drvthis->config_get_string(drvthis->name, "size", 0, "20x4");
  if ((sscanf(s, "%dx%d", &(p->width), &(p->height)) != 2)
      || (p->width <= 0) || (p->width > LCD_MAX_WIDTH)
      || (p->height <= 0) || (p->height > LCD_MAX_HEIGHT)) {
    report(RPT_ERR, "%s: cannot read Size %s", drvthis->name, s);
    return -1;
  }

  p->cellwidth = DEFAULT_CELLWIDTH;
  p->cellheight = DEFAULT_CELLHEIGHT;
  p->contrast = DEFAULT_CONTRAST;
  p->brightness = DEFAULT_BRIGHTNESS;

  p->framebuf = malloc(p->width * p->height);
  if(p->framebuf == NULL) {
    report(RPT_ERR, "unable to allocate frame buffer");
    return -1;
  }

  p->lcdcontents = malloc(p->width * p->height);
  if(p->lcdcontents == NULL) {
    report(RPT_ERR, "unable to allocate lcdcontents store ");
    return -1; 
  }

  p->port = drvthis->config_get_int(drvthis->name, "port", 0, 0x82);

  /* Get serial device */
  strncpy(device, drvthis->config_get_string(drvthis->name, "Device", 0, DEFAULT_DEVICE), sizeof(device));
  device[sizeof(device)-1] = '\0';

  p->fd = open(device, O_RDWR);
  if (p->fd < 0) {
    report(RPT_ERR, "can't open device");
    return -1;
  }

  p->port = p->port >> 1;
  if (ioctl(p->fd, I2C_SLAVE, p->port) < 0) {
    report(RPT_ERR, "can't set slave address");
    return -1;
  }

  bwi2c_clear(drvthis);

  return 0;
}

/**
 * Close the driver (do necessary clean-up). (Required)
 * \param drvthis  Pointer to driver structure.
 */
MODULE_EXPORT void
bwi2c_close(Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;
  
  if(p != NULL) {
    
    if(p->framebuf != NULL) {
      free(p->framebuf);
      p->framebuf = NULL;
    }

    if(p->lcdcontents != NULL) {
      free(p->lcdcontents);
      p->lcdcontents = NULL;
    }

    if (p->fd >= 0) {
      close(p->fd);
    }

    free(p);
  }
  drvthis->store_private_ptr(drvthis, NULL);
}

/**
 * Clear the screen. (Required for output)
 * \param drvthis  Pointer to driver structure.
 */
MODULE_EXPORT void
bwi2c_clear(Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;

  memset(p->framebuf, ' ', p->width * p->height);
}

/**
 * Return the display width in characters. (Required for output)
 * \param drvthis  Pointer to driver structure.
 * \return         Number of characters the display is wide.
 */
MODULE_EXPORT int
bwi2c_width(Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;

  return p->width;
}

/**
 * Return the display height in characters. (Required for output)
 * \param drvthis  Pointer to driver structure.
 * \return         Number of characters the display is high.
 */
MODULE_EXPORT int
bwi2c_height(Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;

  return p->height;
}

/**
 * Print a string on the screen at position (x,y). (Required for output)
 * The upper-left corner is (1,1), the lower-right corner is (p->width, p->height).
 * \param drvthis  Pointer to driver structure.
 * \param x        Horizontal character position (column).
 * \param y        Vertical character position (row).
 * \param string   String that gets written.
 *
 * \see debug_flush
 */
MODULE_EXPORT void
bwi2c_string(Driver *drvthis, int x, int y, const char string[])
{
  PrivateData *p = drvthis->private_data;
  int i;

  report(RPT_INFO, "%s(%i,%i,%.40s)", __FUNCTION__, x, y, string);

  y--;      /* Convert 1-based coords to 0-based... */
  x--;

  if((y < 0) || (y >= p->height)){
    return;
  }

  for(i=0; (string[i] != '\0') && (x < p->width); i++, x++) {
    if(x >= 0) {
      p->framebuf[(y * p->width) + x] = string[i];
    }
  }
}


/**
 * Print a character on the screen at position (x,y). (Required for output)
 * The upper-left corner is (1,1), the lower-right corner is (p->width, p->height).
 * \param drvthis  Pointer to driver structure.
 * \param x        Horizontal character position (column).
 * \param y        Vertical character position (row).
 * \param c        Character that gets written.
 *
 * \note Keep in mind that character 0x00 may be a valid character on your
 * display and it is used by the bignum library. Avoid using string handling
 * functions here!
 *
 *
 */
MODULE_EXPORT void
bwi2c_chr(Driver *drvthis, int x, int y, char c)
{
  PrivateData *p = drvthis->private_data;

  x--;  /* Convert 1-based coords to 0-based... */
  y--;

  if ((x >= 0) && (y >= 0) && (x < p->width) && (y < p->height)) {
    p->framebuf[(y * p->width) + x] = c;
  }
}

MODULE_EXPORT void
bwi2c_flush(Driver *drvthis)
{

  PrivateData *p = (PrivateData *) drvthis->private_data;
  int x, y;
  int wid = p->width;
  char ch;
  int firstdiff;
  int lastdiff;

  /* Update LCD incrementally by comparing with last contents */
  for (y = 0; y < p->height; y++) {
    firstdiff = -1;
    lastdiff = 0;
    for (x = 0; x < wid; x++) {
      ch = p->framebuf[(y * wid) + x];
      if (ch != p->lcdcontents[(y * wid) + x]) {
        p->lcdcontents[(y * wid) + x] = ch;
        if (firstdiff == -1) {
          firstdiff = x;
        }
        lastdiff = x;
      }
    }

    if (firstdiff >= 0) {
      bwi2c_move_cursor(p->fd, p->port, firstdiff, y);
      bwi2c_send_text(p->fd, p->port, p->framebuf + (y * wid) + firstdiff,  lastdiff - firstdiff + 1);
    }
  }
}

/**
 * Get current LCD contrast. (Optional)
 * This is only the locally stored contrast, the contrast value
 * cannot be retrieved from the LCD.
 * \param drvthis  Pointer to driver structure.
 * \return         Stored contrast in promille.
 */
MODULE_EXPORT int
bwi2c_get_contrast(Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;

  return p->contrast;
}

/**
 * Change LCD contrast. (Optional)
 * \param drvthis  Pointer to driver structure.
 * \param promille New contrast value in promille.
 */
MODULE_EXPORT void
bwi2c_set_contrast(Driver *drvthis, int promille)
{
  PrivateData *p = drvthis->private_data;

  bwi2c_set_reg_value8 (p->fd, p->port, 0x12, promille);

  p->contrast = promille;
}

/**
 * Retrieve brightness. (Optional)
 * \param drvthis  Pointer to driver structure.
 * \param state    Brightness state (on/off) for which we want the value (ignored).
 * \return         Stored brightness in promille.
 */
MODULE_EXPORT int
bwi2c_get_brightness(Driver *drvthis, int state)
{
  PrivateData *p = drvthis->private_data;

  return p->brightness;
}

/**
 * Set on/off brightness. (Optional)
 * \param drvthis   Pointer to driver structure.
 * \param state     Brightness state (on/off) for which we want to store the value (ignored).
 * \param promille  New brightness in promille.
 */
MODULE_EXPORT void
bwi2c_set_brightness(Driver *drvthis, int state, int promille)
{
  PrivateData *p = drvthis->private_data;

  bwi2c_set_reg_value8 (p->fd, p->port, 0x13, promille);

  p->brightness = promille;
}

/**
 * Provide some information about this driver. (Optional)
 * \param drvthis  Pointer to driver structure.
 * \return         Constant string with information.
 */
MODULE_EXPORT const char *
bwi2c_get_info(Driver *drvthis)
{
  static char *info_string = "Driver for BitWizard i2c_lcd expansion board, v1.5";

  return info_string;
}

MODULE_EXPORT void 
bwi2c_set_reg_value8 (int fd, int addr, int reg, int val)
{
  char buf[5];
  buf[0] = addr;
  buf[1] = reg;
  buf[2] = val;
  bwi2c_transfer (fd, 3, buf);
}

MODULE_EXPORT void 
bwi2c_transfer (int fd, int len, char *buf)
{
   write (fd, buf+1, len-1);
}

MODULE_EXPORT 
void bwi2c_send_text (int fd, int addr, char *str, int len)
{
  char *buf;
  
  buf = malloc (len + 5);
  buf[0] = addr;
  buf[1] = 0;
  strncpy(buf+2, str, len);
  bwi2c_transfer (fd, len+2, buf);
  free (buf);
}

MODULE_EXPORT
void bwi2c_move_cursor (int fd, int addr, int x, int y)
{
  int c = (y << 5) | x;
  bwi2c_set_reg_value8(fd, addr, 0x11, c);
}